﻿using HPPH;
using HPPH.System.Drawing;
using StableDiffusion.NET;
using System.Data.SQLite;

namespace LLM_Doghouse
{
    public partial class AimGenForm : Form
    {
        private string mdlPath = "";
        private string vaePath = "";
        private string embPath = "";
        private string DatabasePath = "";
        public AimGenForm(string DatabasePath)
        {
            InitializeComponent();
            this.DatabasePath = DatabasePath;
        }


        private void AimGenForm_Load(object sender, EventArgs e)
        {
            InitSetting();
            txtMdl.Text = mdlPath;
            txtVae.Text = vaePath;
            txtEmb.Text = embPath;
        }

        public void InitSetting()
        {
            try
            {
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    cmd.CommandText = $"select * from sd where no = 1";
                    using (var reader = cmd.ExecuteReader())
                    {
                        if (reader.Read())
                        {
                            mdlPath = reader["mdl"].ToString();
                            vaePath = reader["vae"].ToString();
                            embPath = reader["emb"].ToString();
                        }
                    }
                }
                con.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void txtMdl_Enter(object sender, EventArgs e)
        {
            txtMdl.SelectAll();
        }

        private void txtVae_Enter(object sender, EventArgs e)
        {
            txtVae.SelectAll();
        }

        private void txtEmb_Enter(object sender, EventArgs e)
        {
            txtEmb.SelectAll();
        }

        private void btnUpd_Click(object sender, EventArgs e)
        {
            mdlPath = txtMdl.Text;
            vaePath = txtVae.Text;
            embPath = txtEmb.Text;
            if (CheckItem()) return;
            if (MessageBox.Show("Update record. Are you sure?", this.Text, MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.No) return;
            string strSql = "update sd set " +
                            $"mdl='{txtMdl.Text}'," +
                            $"vae='{txtVae.Text}'," +
                            $"emb='{txtEmb.Text}'" +
                            $" where no = 1";
            this.DialogResult = DialogResult.OK;
            if (ExcSql(strSql)) this.Close();
        }

        private void btnMdl_Click(object sender, EventArgs e)
        {
            string strFolder = "";
            if (txtMdl.Text.ToString() != null) strFolder = Path.GetDirectoryName(txtMdl.Text.ToString());
            string strRtn = OpenFile(strFolder: strFolder, "Open Model File", txtMdl.Text.ToString());
            if (strRtn != "") txtMdl.Text = strRtn;
        }

        private void btnVae_Click(object sender, EventArgs e)
        {
            string strFolder = "";
            if (txtVae.Text.ToString() != null) strFolder = Path.GetDirectoryName(txtVae.Text.ToString());
            string strRtn = OpenFile(strFolder, "Open Model File", txtVae.Text.ToString());
            if (strRtn != "") txtVae.Text = strRtn;
        }

        private void btnEmb_Click(object sender, EventArgs e)
        {
            string strFolder = "";
            if (txtEmb.Text.ToString() != null) strFolder = Path.GetDirectoryName(txtEmb.Text.ToString());
            string strRtn = OpenFile(strFolder, "Open Model File", txtEmb.Text.ToString());
            if (strRtn != "") txtEmb.Text = strRtn;
        }

        private string OpenFile(string strFolder, string strTitle, string strInitFile)
        {
            OpenFileDialog dialog = new OpenFileDialog();
            dialog.FileName = strInitFile;
            dialog.InitialDirectory = strFolder;
            dialog.FilterIndex = 2;
            dialog.Title = strTitle;
            dialog.RestoreDirectory = true;
            dialog.CheckFileExists = true;
            dialog.CheckPathExists = true;
            dialog.Filter = $"model file|*.gguf;*.bin;*.safetensors;*.pt";
            if (dialog.ShowDialog() == DialogResult.OK) return (dialog.FileName);
            return ("");
        }

        private bool ExcSql(string strSql)
        {
            try
            {
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    cmd.CommandText = strSql;
                    cmd.ExecuteNonQuery();
                    con.Close();
                }
                return true;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
        }

        public bool CheckItem(bool mb = true)
        {
            if (mdlPath.Trim() == "")
            {
                if(mb) MessageBox.Show("Stable Diffusion Model file does not exist.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return (true);
            }
            if (mdlPath.Trim() != "")
            {
                if (!File.Exists(mdlPath))
                {
                    if (mb) MessageBox.Show("Stable Diffusion Model file does not exist.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return (true);
                }
            }
            if (vaePath.Trim() != "")
            {
                if (!File.Exists(vaePath))
                {
                    if (mb) MessageBox.Show("Vae file does not exist.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return (true);
                }
            }
            if (embPath.Trim() != "")
            {
                if (!File.Exists(embPath))
                {
                    if (mb) MessageBox.Show("Embedding file does not exist.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return (true);
                }
            }

            return (false);
        }

        public void AimGenRun(string prompt, string imgPath)
        {
            const int intWidth = 256;
            const int intHeight = 256;
            const float fltScale = 11f;
            const int intSteps = 12;
            IImage objImage;
            try
            {
                if (CheckItem()) return;
                //Load Model
                DiffusionModel sdModel;
                if (vaePath != "" && embPath != "")
                    sdModel = ModelBuilder.StableDiffusion(mdlPath).WithMultithreading().WithVae(vaePath).WithSchedule(Schedule.Karras).WithEmbeddingSupport(embPath).Build();
                else if (vaePath != "" && embPath == "")
                    sdModel = ModelBuilder.StableDiffusion(mdlPath).WithMultithreading().WithVae(vaePath).WithSchedule(Schedule.Karras).Build();
                else if (vaePath == "" && embPath != "")
                    sdModel = ModelBuilder.StableDiffusion(mdlPath).WithMultithreading().WithSchedule(Schedule.Karras).WithEmbeddingSupport(embPath).Build();
                else
                    sdModel = ModelBuilder.StableDiffusion(mdlPath).WithMultithreading().WithSchedule(Schedule.Karras).Build();

                string antiprompt = "(worst quality:2) , (low quality:2) , (normal quality:2) , lowres, ugly face, unclear eyes, bad mouth, bad tooth, bad anatomy, extra legs, (bad fingers, bad hands, missing fingers), negative_hand-neg";
                long lngSeed = 0;

                //Seed Random
                Random rSeed = new Random();
                lngSeed = rSeed.Next(0, int.MaxValue);
                DiffusionParameter parameter = sdModel.GetDefaultParameter()
                                                        .WithNegativePrompt(antiprompt)
                                                        .WithSize(intWidth, intHeight)
                                                        .WithSteps(intSteps)
                                                        .WithSeed(lngSeed)
                                                        .WithSampler(Sampler.DPMPP2Mv2)
                                                        .WithCfg(fltScale);

                objImage = sdModel.TextToImage(prompt, parameter);
                File.WriteAllBytes(imgPath, objImage.ToPng());

                var frm = new PreviewForm();
                frm.picPreview.ImageLocation = imgPath;
                frm.Show();

            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

    }
}
