﻿using LLama.Common;
using LLama;
using System.Data.SQLite;
using LLama.Sampling;

namespace LLM_Doghouse
{
    internal class HistorySummary
    {
        private string DatabasePath = "";
        public HistorySummary(string DatabasePath) 
        {
            this.DatabasePath = DatabasePath;
        }

        public async void Run(uint HistoryMax, string strTable)
        {
            try
            {
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    cmd.CommandText = $"select * from {strTable} where sq=1";
                    using (var reader = cmd.ExecuteReader())
                    {
                        reader.Read();
                        reader.Close();
                    }
                }
                //DBクローズ
                con.Close();
            }
            catch (Exception ex) 
            {
                string msg = ex.Message;    
                return;
            }

            try
            {
                string strChtHis = ""; // ChatDBから行を読み込む
                int i = 0;
                uint intCtx = 0;      //一度に処理できる文字数
                float fltTmp = 0;
                int intGpu = 0;
                string strMdl = "";
                long svSq = 0;      //処理を開始したSq
                long svSqMax = 0;   //処理を終了したSq
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    //LLMのパラメータ取得
                    cmd.CommandText = $"select * from sum where no=1";
                    using (var reader = cmd.ExecuteReader())
                    {
                        if (reader.Read())
                        {
                            intCtx = uint.Parse(reader["ctx"].ToString());
                            fltTmp = float.Parse(reader["tmp"].ToString());
                            intGpu = int.Parse(reader["gpu"].ToString());
                            strMdl = reader["mdl"].ToString();
                        }
                        reader.Close();
                    }
                    //有効文字数をカウント
                    cmd.CommandText = $"select sum(length(msg))  from {strTable} where flg=0";
                    using (var reader = cmd.ExecuteReader())
                    {
                        reader.Read();
                        //Console.WriteLine($"文字数：{reader[0]}");
                        //チャット履歴の最大行に満たないか、最大文字数を越えても処理できる文字数が充分でない場合は処理をしない
                        if ((long)reader[0] < (long)HistoryMax || (long)reader[0] - (long)HistoryMax < (long)intCtx)
                        {
                            con.Close();
                            return;
                        }
                        reader.Close();
                    }
                    //有効行を呼んで要約する
                    cmd.CommandText = $"select * from {strTable} where flg=0 and sq>3 order by sq";
                    string strCht = "";
                    using (var reader = cmd.ExecuteReader())
                    {
                        while (reader.Read())
                        {
                            strCht = (string)reader["id"] + ": " + ((string)reader["msg"]).Replace("User:", "").Replace("Assistant:", "").Replace("assistant:", "").Trim() + "\n";
                            i += (int)strCht.Length;
                            if (i > intCtx) { break; }
                            strChtHis += strCht;
                            //システム行以外で処理を開始した行を記録
                            if ((string)reader["id"] != "System" && svSq == 0)
                            {
                                svSq = (long)reader["sq"];
                            }
                            svSqMax = (long)reader["sq"];
                        }
                        reader.Close();
                    }
                }


                //LLMモデルのロードとパラメータの設定
                ModelParams modPara = new(strMdl)
                {
                    ContextSize = intCtx,
                    GpuLayerCount = intGpu
                };

                ChatHistory chtHis = new ChatHistory();
                chtHis.AddMessage(AuthorRole.System, "#命令書\n" +
                                                    "・あなたは優秀な編集者です。\n" +
                                                    "・ユーザーとアシスタントの会話を要約してください。\n" +
                                                    "#条件\n" +
                                                    "・重要なキーワードを取りこぼさない。\n" +
                                                    "#出力形式\n" +
                                                    "・例）要約: ふたりは親密な会話しました。");
                using LLamaWeights llmWeit = LLamaWeights.LoadFromFile(modPara);
                using LLamaContext llmContx = llmWeit.CreateContext(modPara);
                InteractiveExecutor itrEx = new(llmContx);
                ChatSession chtSess = new(itrEx, chtHis);
                var varHidewd = new LLamaTransforms.KeywordTextOutputStreamTransform(["User:", "Assistant:"]);
                chtSess.WithOutputTransform(varHidewd);
                InferenceParams infPara = new()
                {
                    SamplingPipeline = new DefaultSamplingPipeline()
                    {
                        Temperature = fltTmp,
                        Seed = 1337,
                    },
                    AntiPrompts = new List<string> { "User:" }
                };

                // ユーザーのターン
                string strInput =  strChtHis;
                ChatHistory.Message msg = new(AuthorRole.User, strInput);

                // ＡＩのターン
                string strMsg = "";
                await foreach (string strAns in chtSess.ChatAsync(msg, infPara))
                {
                    strMsg += strAns;
                }

                using (var cmd = new SQLiteCommand(con))
                {
                    //sq=1に要約内容を更新する(要約という文字、改行、空白を削除）
                    string strSql = $"update {strTable} set msg= '{strMsg.Replace("要約:", "").Replace("\r", "").Replace("\n", "").Trim()}' where sq=2";
                    cmd.CommandText = strSql;
                    cmd.ExecuteNonQuery();

                    //処理した行を論理削除する
                    strSql = $"update {strTable} set flg=1 where (sq between {svSq} and {svSqMax}) and id<>'System' and flg=0";
                    cmd.CommandText = strSql;
                    cmd.ExecuteNonQuery();
                }
                //DBクローズ
                con.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "HistorySummary", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }


        }

    }
}
