﻿using LLama;
using LLama.Common;
using LLama.Native;
using LLama.Sampling;
using System.Text.RegularExpressions;
using System.Data.SQLite;

namespace LLM_Doghouse
{
    internal class Llava
    {
        public string Message { get; set; } = "";
        private string strMsg = "";
        private string DatabasePath = "";
        public Llava(string DatabasePath)
        {
            this.DatabasePath = DatabasePath;
        }

        public async ValueTask<string> Run(string imgPath)
        {
            uint intCtx = 0;      //一度に処理できる文字数
            float fltTmp = 0;
            int intGpu = 0;
            string strMdl = "";
            string strClp = "";
            try
            {
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    //LLMのパラメータ取得
                    cmd.CommandText = $"select * from llv where no=1";
                    using (var reader = cmd.ExecuteReader())
                    {
                        if (reader.Read())
                        {
                            intCtx = uint.Parse(reader["ctx"].ToString());
                            fltTmp = float.Parse(reader["tmp"].ToString());
                            intGpu = int.Parse(reader["gpu"].ToString());
                            strMdl = reader["mdl"].ToString();
                            strClp = reader["clp"].ToString();
                        }
                        reader.Close();
                    }
                }
                con.Close();

                var prompt = $"{{{imgPath}}}\nUSER:\n画像に何が写っているか教えてください。\nASSISTANT:\n";
                var modPara = new ModelParams(strMdl)
                {
                    ContextSize = intCtx,
                    GpuLayerCount = intGpu,
                };
                var llmWeit = LLamaWeights.LoadFromFile(modPara);
                var llmContx = llmWeit.CreateContext(modPara);
                // Llava Init
                using var clipModel = LLavaWeights.LoadFromFile(strClp);
                var itrEx = new InteractiveExecutor(llmContx, clipModel);
                var infPara = new InferenceParams()
                {
                    SamplingPipeline = new DefaultSamplingPipeline()
                    {
                        Temperature = fltTmp,
                        Seed = 1337,
                    },
                    AntiPrompts = ["User:"],
                    MaxTokens = 256
                };

                // Evaluate if we have images
                var imageMatches = Regex.Matches(prompt, "{([^}]*)}").Select(m => m.Value);
                var imageCount = imageMatches.Count();
                var hasImages = imageCount > 0;

                if (hasImages)
                {
                    var imagePathsWithCurlyBraces = Regex.Matches(prompt, "{([^}]*)}").Select(m => m.Value);
                    var imagePaths = Regex.Matches(prompt, "{([^}]*)}").Select(m => m.Groups[1].Value).ToList();

                    itrEx.Context.NativeHandle.KvCacheRemove(LLamaSeqId.Zero, -1, -1);
                    int index = 0;
                    foreach (var path in imagePathsWithCurlyBraces)
                    {
                        // 最初の画像をタグ<image>に置き換え、残りの画像ではタグを削除します。
                        prompt = prompt.Replace(path, index++ == 0 ? "<image>" : "");
                    }
                    // Initialize Images in executor
                    foreach (var image in imagePaths)
                    {
                        itrEx.Images.Add(File.ReadAllBytes(image));
                    }
                }
                strMsg = "";
                await foreach (var text in itrEx.InferAsync(prompt, infPara))
                {
                    strMsg += text;
                }
                return strMsg;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Llava", MessageBoxButtons.OK, MessageBoxIcon.Error);
                strMsg = "";
                return "";
            }
        }

    }
}
