﻿using LLama.Common;
using LLama.Native;
using LLama;
using System.Data;
using System.Data.SQLite;
using System.Text.RegularExpressions;
using LLama.Sampling;

namespace LLM_Doghouse
{
    public partial class LlavaForm : Form
    {
        public LlavaForm(string DatabasePath)
        {
            InitializeComponent();
            this.DatabasePath = DatabasePath;
        }
        private string DatabasePath;
        private uint intCtx = 0;
        private float fltTmp = 0;
        private int intGpu = 0;
        private string strMdl = "";
        private string strClp = "";

        public bool CheckItem(bool mb = true)
        {
            //パラメータチェック
            if (strMdl.IndexOf("\'")>0 || strMdl.IndexOf("\"")>0)
            {
                if (mb) MessageBox.Show("LLM Path contains characters that cannot be used.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return (true);
            }

            if (!File.Exists(strMdl))
            {
                if (mb) MessageBox.Show("LLM Model file does not exist.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return (true);
            }
            if (!File.Exists(strClp))
            {
                if (mb) MessageBox.Show("Clip Model file does not exist.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return (true);
            }
            return (false);
        }

        private void btnMdl_Click(object sender, EventArgs e)
        {
            string strFolder = "";
            if (txtMdl.Text.ToString() != null) strFolder = Path.GetDirectoryName(txtMdl.Text.ToString());
            string strRtn = OpenFile(strFolder: strFolder, "Open Model File", txtMdl.Text.ToString());
            if (strRtn != "") txtMdl.Text = strRtn;
        }

        private void btnUpd_Click(object sender, EventArgs e)
        {

            if (!uint.TryParse(txtCtx.Text, out intCtx)) intCtx = 0;
            if (intCtx == 0) intCtx = 512;
            if (!float.TryParse(txtTmp.Text, out fltTmp)) fltTmp = 0.0f;
            if (!int.TryParse(txtGpu.Text, out intGpu)) intGpu = 0;
            if (intGpu == 0) intGpu = 5;
            strMdl = txtMdl.Text;
            strClp = txtClp.Text;
            txtCtx.Text = intCtx.ToString();
            txtTmp.Text = fltTmp.ToString("0.0");
            txtGpu.Text = intGpu.ToString();

            if (CheckItem()) return;
            if (MessageBox.Show("Update record. Are you sure?", this.Text, MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.No) return;

            string strSql = "update llv set " +
                            $"tmp={fltTmp}," +
                            $"gpu={intGpu}," +
                            $"ctx={intCtx}," +
                            $"mdl='{strMdl}'," +
                            $"clp='{strClp}'" +
                            $" where no=1";
            this.DialogResult = DialogResult.OK;
            if (ExcSql(strSql)) this.Close();
        }

        private void DefineTable()
        {
            try
            {
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    cmd.CommandText = $"CREATE TABLE IF NOT EXISTS \"llv\" (" +
                        "\"no\"  INTEGER NOT NULL DEFAULT 0," +
                        "\"tmp\" REAL NOT NULL DEFAULT 0," +
                        "\"gpu\" REAL NOT NULL DEFAULT 0," +
                        "\"ctx\" REAL NOT NULL DEFAULT 0," +
                        "\"mdl\" TEXT," +
                        "\"clp\" TEXT," +
                        "PRIMARY KEY(\"no\"))";
                    cmd.ExecuteNonQuery();

                    cmd.CommandText = $"select count(*) from llv Where no=1";
                    using (var reader = cmd.ExecuteReader())
                    {
                        long reccount = 0;
                        if (reader.Read()) reccount = (long)reader[0];
                        reader.Close();
                        if (reccount==0)
                        {
                            cmd.CommandText = $"insert into llv(tmp,gpu,ctx,mdl,clp) values(0.0,15,512,'','')";
                            cmd.ExecuteNonQuery();
                        }
                    }
                    con.Close();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "ChatHistoryDB", MessageBoxButtons.OK, MessageBoxIcon.Error);
                //Console.WriteLine(ex.Message);
            }
        }


        private bool ExcSql(string strSql)
        {
            try
            {
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    cmd.CommandText = strSql;
                    cmd.ExecuteNonQuery();
                    con.Close();
                }
                return true;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
        }

        private string OpenFile(string strFolder, string strTitle, string strInitFile)
        {
            OpenFileDialog dialog = new OpenFileDialog();
            dialog.FileName = strInitFile;
            dialog.InitialDirectory = strFolder;
            dialog.FilterIndex = 2;
            dialog.Title = strTitle;
            dialog.RestoreDirectory = true;
            dialog.CheckFileExists = true;
            dialog.CheckPathExists = true;
            dialog.Filter = $"model file|*.gguf;*.bin;*.safetensors";
            if (dialog.ShowDialog() == DialogResult.OK) return (dialog.FileName);
            return ("");
        }

        private void txtCtx_Enter(object sender, EventArgs e)
        {
            txtCtx.SelectAll();
        }

        private void txtTmp_Enter(object sender, EventArgs e)
        {
            txtTmp.SelectAll();
        }

        private void txtGpu_Enter(object sender, EventArgs e)
        {
            txtGpu.SelectAll();
        }

        private void txtMdl_Enter(object sender, EventArgs e)
        {
            txtMdl.SelectAll();
        }

        private void EditSummaryForm_Load(object sender, EventArgs e)
        {
            InitSetting();
            txtCtx.Text = intCtx.ToString();
            txtTmp.Text = fltTmp.ToString("0.0");
            txtGpu.Text = intGpu.ToString();
            txtMdl.Text = strMdl;
            txtClp.Text = strClp;
        }

        public void InitSetting()
        {
            var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
            var con = new SQLiteConnection(conSb.ToString());
            con.Open();
            using (var cmd = new SQLiteCommand(con))
            {
                cmd.CommandText = $"CREATE TABLE IF NOT EXISTS \"llv\" (" +
                    "\"no\"  INTEGER NOT NULL DEFAULT 0," +
                    "\"tmp\" REAL NOT NULL DEFAULT 0," +
                    "\"gpu\" REAL NOT NULL DEFAULT 0," +
                    "\"ctx\" REAL NOT NULL DEFAULT 0," +
                    "\"mdl\" TEXT," +
                    "\"clp\" TEXT," +
                    "PRIMARY KEY(\"no\"))";
                cmd.ExecuteNonQuery();

                cmd.CommandText = $"select count(*) from llv Where no=1";
                using (var reader = cmd.ExecuteReader())
                {
                    long reccount = 0;
                    if (reader.Read()) reccount = (long)reader[0];
                    reader.Close();
                    if (reccount==0)
                    {
                        cmd.CommandText = $"insert into llv(tmp,gpu,ctx,mdl,clp) values(0.0,15,512,'','')";
                        cmd.ExecuteNonQuery();
                    }
                }
                cmd.CommandText = $"select * from llv where no=1";
                using (var reader = cmd.ExecuteReader())
                {
                    if (reader.Read())
                    {
                        intCtx = uint.Parse(reader["ctx"].ToString());
                        fltTmp = float.Parse(reader["tmp"].ToString());
                        intGpu = int.Parse(reader["gpu"].ToString());
                        strMdl = reader["mdl"].ToString();
                        strClp = reader["clp"].ToString();
                    }
                }
            }
            con.Close();
        }

        private void txtClp_Enter(object sender, EventArgs e)
        {
            txtClp.SelectAll();
        }

        private void btnClp_Click(object sender, EventArgs e)
        {
            string strFolder = "";
            if (txtClp.Text.ToString() != null) strFolder = Path.GetDirectoryName(txtClp.Text.ToString());
            string strRtn = OpenFile(strFolder: strFolder, "Open Model File", txtClp.Text.ToString());
            if (strRtn != "") txtClp.Text = strRtn;
        }

        private void LlavaForm_Load(object sender, EventArgs e)
        {
            InitSetting();
            txtCtx.Text = intCtx.ToString();
            txtTmp.Text = fltTmp.ToString("0.0");
            txtGpu.Text = intGpu.ToString();
            txtMdl.Text = strMdl;
            txtClp.Text = strClp;
        }
    }
}
