﻿using LLama.Common;
using LLama.Sampling;
using LLama;
using System.Data;
using System.Data.SQLite;
using System.Diagnostics;
using System.Media;
using Windows.Media.SpeechRecognition;
using System.Text.RegularExpressions;
using DSharpPlus;
using DSharpPlus.Entities;
using DSharpPlus.VoiceNext;
using DSharpPlus.EventArgs;

namespace LLM_Doghouse
{
    public partial class MainForm : Form
    {
        private string DatabasePath = Application.StartupPath + @"\ChatDB.db";
        private DataGridViewCellStyle dvwStyle1 = new DataGridViewCellStyle();
        private DataGridViewCellStyle dvwStyle2 = new DataGridViewCellStyle();

        private bool Op = true;             //本番フラグ
        private int silenceTimerCnt = 0;    //Timerカウント
        private bool Wt = false;            //待ちフラグ
        private ChatHistoryDB chtDB;       //チャットログシステム
        private ModelParams modPara;
        private LLamaWeights llmWeit;
        private LLamaContext llmContx;
        private InteractiveExecutor? itrEx;
        private ChatHistory chtHis;
        private ChatSession chtSess;
        private LLamaTransforms.KeywordTextOutputStreamTransform? llmHidewd;
        private InferenceParams infPara;
        private SpeechRecognizer recognizer;
        private int intSyt = 0;
        private string strCvo = "";
        private uint intVvx = 0;
        private bool Sd = false;
        private bool Sm = false;
        private bool Lv = false;
        private string Sdprompt = "";
        private DataTable dtChat;
        private bool booBot = false;
        private string strTkn = "";
        private ulong lngCid = 0;
        //Discord Bot変数
        private DiscordConfiguration dcconfig;
        private DiscordClient dcclient;
        private DiscordActivity dcstatus;
        private VoiceNextConnection dcvoiceCon;
        //VOICEVOX
        private Core core;
        //Emotion Image
        private string[] strDefPath = { "imgUsu.png", "imgSml.png", "imgAfc.png", "imgSkp.png", "imgCnc.png", "imgNeg.png" };
        private string[] strEmtPath = { "imgUsu.png", "imgSml.png", "imgAfc.png", "imgSkp.png", "imgCnc.png", "imgNeg.png" };
        private Image[] imgEmt;

        public MainForm()
        {
            InitializeComponent();
        }

        private async void MainForm_Load(object sender, EventArgs e)
        {
            if (Properties.Settings.Default.WindowSize.Width == 0) Properties.Settings.Default.Upgrade();
            this.Location = Properties.Settings.Default.WindowLocation;
            this.Size = Properties.Settings.Default.WindowSize;

            txtMsg.Enabled=false;
            chkMic.Enabled=false;
            lblInfo.Visible = false;
            picDc.Visible = false;

            try
            {
                cboChatList.DataSource = mng();
                if (cboChatList.Items.Count==0) btnLoad.Enabled = false;

                Font dvwFont = new Font("ＭＳ ゴシック", 12);
                dvwStyle1.BackColor = Color.FromArgb(60, 60, 60);
                dvwStyle1.ForeColor = Color.FromArgb(255, 255, 192);
                dvwStyle1.Font = dvwFont;
                dvwStyle2.BackColor = Color.Gray;
                dvwStyle2.ForeColor = Color.White;
                dvwStyle2.Font = dvwFont;
                // SpeechRecognitionの設定
                recognizer = new SpeechRecognizer();
                // Set timeout settings.
                recognizer.Timeouts.InitialSilenceTimeout = TimeSpan.FromSeconds(6.0); // (認識結果が生成されるまでの) 無音を検出し、音声入力が続かないと見なす時間の長さ。
                recognizer.Timeouts.BabbleTimeout = TimeSpan.FromSeconds(4.0);         //認識できないサウンド (雑音) のリッスンを継続し、音声入力が終了したと見なし、認識処理を終了するまでの時間の長さ。
                recognizer.Timeouts.EndSilenceTimeout = TimeSpan.FromSeconds(1.2);     //(認識結果が生成された後の) 無音を検出し、音声入力が終了したと見なす時間の長さ。
                await recognizer.CompileConstraintsAsync();

                recognizer.ContinuousRecognitionSession.ResultGenerated += ContinuousRecognitionSession_ResultGenerated;
                //AI会話タイムアウト
                recognizer.ContinuousRecognitionSession.Completed += ContinuousRecognitionSession_Completed;
                //GPU使用メモリ編集
                GPU_UsageMemory(); 
                //要約環境設定チェック
                CheckSm();
                //Stable Diffusion設定チェック
                CheckSd();
                //Llva設定チェック
                CheckLv();
                if(cboChatList.Items.Count > 0)
                {
                    //引数起動の場合はロードを行う
                    string[] cmds = System.Environment.GetCommandLineArgs();
                    //コマンドライン引数を列挙する
                    int intNo = 0;
                    foreach (string cmd in cmds)
                    {
                        if (cmd.StartsWith("/ch"))
                        {
                            if (int.TryParse(cmd.Replace("/ch", ""), out intNo))
                            {
                                if (intNo > 0)
                                {
                                    cboChatList.SelectedValue = intNo;
                                    if (cboChatList.SelectedItem == null) break;
                                    ValueTask<bool> booTask = LoadLLM();
                                    booTask.AsTask().Wait();
                                    if (!booTask.Result) LoadSetting();
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                Application.Exit();
            }
        }
        private void MainForm_SizeChanged(object sender, EventArgs e)
        {
            if (dvwMsg.Rows.Count>0) dvwMsg.Columns[1].Width = dvwMsg.Width - 90;
        }
        private void btnLoad_Click(object sender, EventArgs e)
        {
            ValueTask<bool> booTask = LoadLLM();
            booTask.AsTask().Wait();
            if (booTask.Result) return;
            LoadSetting();
        }
        private ValueTask<bool> LoadLLM()
        {
            //ロードLLM
            string strMdl = "";
            float fltTmp = 0;
            int intGpu = 0;
            uint intCtx = 0;
            string strSys = "";
            strEmtPath = strDefPath;
            imgEmt = new Image[strEmtPath.Length];

            try
            {
                //初期化処理
                ClearItems();
                //DBからパラメータ読み込み
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    cmd.CommandText = $"select * from mng where no={cboChatList.SelectedValue}";
                    using (var reader = cmd.ExecuteReader())
                    {
                        if (reader.Read())
                        {
                            strMdl = reader["mdl"].ToString();
                            fltTmp = float.Parse(reader["tmp"].ToString());
                            intGpu = int.Parse(reader["gpu"].ToString());
                            intCtx = uint.Parse(reader["ctx"].ToString());
                            strSys = reader["sys"].ToString();
                            intSyt = int.Parse(reader["syt"].ToString());
                            strCvo = reader["cvo"].ToString();
                            intVvx = uint.Parse(reader["vvx"].ToString());
                            if (reader.FieldCount > 10)
                            {
                                booBot = reader["bot"].ToString() == "1" ? true : false;
                                strTkn = reader["tkn"].ToString();
                                lngCid = ulong.Parse(reader["cid"].ToString());
                                if (reader.FieldCount > 13)
                                {
                                    for (int i = 0; i < strEmtPath.Length; i++)
                                    {
                                        if (reader.GetString(i+13) !="") strEmtPath[i] = reader.GetString(i+13);
                                    }
                                }
                            }
                            else
                            {
                                booBot = false;
                                strTkn = "";
                                lngCid = 0;
                            }
                            if (intSyt == 1) { lblSyt.Text = strCvo; }
                            if (intSyt == 2) { lblSyt.Text = vvx_name(intVvx); }

                            for(int i = 0;i < strEmtPath.Length; i++) 
                            {
                                if(strEmtPath[i] != "")
                                {
                                    if(File.Exists(strEmtPath[i]))
                                    {
                                        imgEmt[i] = Image.FromFile(strEmtPath[i]);
                                    }
                                }
                            }
                            picEmt.Image = imgEmt[0];
                        }
                        else
                        {
                            MessageBox.Show("Management information cannot be loaded.", "LLM Doghouse", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            con.Close();
                            return ValueTask.FromResult<bool>(true);
                        }
                    }
                }
                con.Close();

                string strTable = $"ch{cboChatList.SelectedValue}";

                //パラメータチェック
                if (!File.Exists(strMdl))
                {
                    MessageBox.Show("LLM Model file does not exist.", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return ValueTask.FromResult<bool>(true);
                }
                if (intGpu == 0) intGpu = 5;
                if (intCtx == 0) intCtx = 512;
                //チャットログ要約処理
                if (Sm)
                {
                    lblInfo.Text = "Now summarizing the history...";
                    lblInfo.Visible = true;
                    Application.DoEvents();
                    var sum = new HistorySummary(DatabasePath);
                    sum.Run(intCtx / 2, strTable);
                    sum = null;
                }
                //LLMモデルのロードとパラメータの設定
                lblInfo.Text = "Now loading the history...";
                lblInfo.Visible = true;
                Application.DoEvents();
                modPara = new ModelParams(strMdl)
                {
                    ContextSize = intCtx,
                    GpuLayerCount = intGpu,
                };
                llmWeit = LLamaWeights.LoadFromFile(modPara);
                llmContx = llmWeit.CreateContext(modPara);
                itrEx = new(llmContx);
                //チャットログを読み込みます。
                chtHis = new ChatHistory();
                chtDB = new ChatHistoryDB(DatabasePath, chtHis, strTable, strSys);
                chtSess = new(itrEx, chtHis);
                llmHidewd = new LLamaTransforms.KeywordTextOutputStreamTransform(["User: ", "Assistant: "]);
                chtSess.WithOutputTransform(llmHidewd);
                infPara = new InferenceParams()
                {
                    SamplingPipeline = new DefaultSamplingPipeline()
                    {
                        Temperature = fltTmp,
                        Seed = 1337,
                    },
                    AntiPrompts = ["User:"],
                    MaxTokens = 256,
                };
                //VOICEVOX Coreをここで読み込んでおく
                InitVoicevox(intVvx);
                //GPU使用メモリ編集
                GPU_UsageMemory();
                //Discord Bot Enabledは、Discord Botを起動する
                if (booBot) DiscordBot();

                lblInfo.Visible = false;
                txtMsg.ReadOnly = false;
                txtMsg.Focus();
                return ValueTask.FromResult<bool>(false);
            }
            catch (Exception ex)
            {
                lblInfo.Visible = false;
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return ValueTask.FromResult<bool>(true);
            }
        }
        private async void ClearItems()
        {
            //初期化処理
            chtDB = null;
            modPara = null;
            if (llmWeit != null) llmWeit.Dispose();
            llmWeit = null;
            if (llmContx != null) llmContx.Dispose();
            llmContx = null;
            itrEx = null;
            chtHis = null;
            chtSess = null;
            llmHidewd = null;
            infPara = null;
            if(core != null) core.Dispose();
            core = null;

            if(booBot)
            {
                //Discord Event Stop & Clear
                dcclient.MessageCreated -= OnMessageCreated;
                dcclient.Ready -= OnClientReady;
                dcclient.GuildAvailable -= OnGuildAvailable;
                dcclient.VoiceStateUpdated -= OnVoiceStateUpdated;
                if(dcvoiceCon is not null) dcvoiceCon.Disconnect();
                await dcclient.DisconnectAsync();
                dcclient.Dispose();
                picDc.Visible = false;
            }
        }
        private async Task Chat(string strUserInput)
        {
            try
            {
                if (!Wt)
                {
                    Wt = true;
                    ChatHistory.Message msgText = new(AuthorRole.User, strUserInput);
                    if (Op) chtDB.WriteHistory(AuthorRole.User, "User: "+ strUserInput);
                    WriteHistory(AuthorRole.User, strUserInput);
                    Invoke(dvwMsg_Refresh);
                    // 回答の表示
                    string strMsg = "";
                    await foreach (string strText in chtSess.ChatAsync(msgText, infPara)) strMsg += strText;
                    //発信するときは「User:」や「Assistant:」を抜く
                    string strSndmsg = strMsg.Replace("User:", "").Replace("Assistant:", "").Replace("assistant:", "").Trim();
                    //画像生成キーワード探索
                    if (strMsg.IndexOf("[")  != -1)
                    {
                        var rgxRtn = new Regex(@"\[(.+?)\]").Matches(strMsg);
                        if (rgxRtn.Count > 0)
                        {
                            Sdprompt = rgxRtn[0].Value.Replace("[", "").Replace("]", "").Trim();
                        }
                    }
                    if (Op) chtDB.WriteHistory(AuthorRole.Assistant, strMsg);
                    WriteHistory(AuthorRole.Assistant, strSndmsg);
                    Invoke(dvwMsg_Refresh);
                    if (chkMic.Checked) 
                    {
                        SpeechSynthesis(strSndmsg); //音声ソフト実行
                        Thread.Sleep(3000);
                    }
                    Wt = false;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }
        private async void ContinuousRecognitionSession_ResultGenerated(SpeechContinuousRecognitionSession sender, SpeechContinuousRecognitionResultGeneratedEventArgs e)
        {
            //AI会話
            if (!Wt && chkMic.Checked)
            {
                silenceTimerCnt = 0;
                silenceTimer.Stop();
                await Chat(e.Result.Text.Replace("MK", "マンコ").Replace("TK", "チンコ").Replace("TP", "チンポ"));
                silenceTimer.Start();
            }
        }
        private async void ContinuousRecognitionSession_Completed(SpeechContinuousRecognitionSession sender, SpeechContinuousRecognitionCompletedEventArgs e)
        {
            if (chkMic.Checked)
            {
                if (recognizer.State == SpeechRecognizerState.Idle)
                {
                    await recognizer.ContinuousRecognitionSession.StartAsync();
                }
            }
        }
        public MemoryStream CevioAI(string strMsg, string strCast, bool booPlay = true)
        {
            try
            {
                var ms = new MemoryStream();
                string strWavPath = Application.StartupPath + @"\output.wav";
                dynamic service = Activator.CreateInstance(Type.GetTypeFromProgID("CeVIO.Talk.RemoteService2.ServiceControl2V40"));
                service.StartHost(false);
                dynamic talker = Activator.CreateInstance(Type.GetTypeFromProgID("CeVIO.Talk.RemoteService2.Talker2V40"));
                talker.Cast = strCast;
                if (booPlay)
                {
                    dynamic result = talker.Speak(strMsg);
                    result.Wait();
                    ms.Dispose();
                    ms = null;
                }
                else
                {
                    dynamic result = talker.OutputWaveToFile(strMsg, strWavPath);
                    using (FileStream file = new FileStream(strWavPath, FileMode.Open, FileAccess.Read))
                    file.CopyTo(ms);
                }
                //開放忘れるとメモリリーク
                System.Runtime.InteropServices.Marshal.ReleaseComObject(talker);
                System.Runtime.InteropServices.Marshal.ReleaseComObject(service);
                return ms;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }
        }

        public void InitVoicevox(uint intSpeaker)
        {
            MainForm me = new MainForm();
            core = me.GetDefaultCoreObject();
            core.LoadModel(intSpeaker);
        }

        public MemoryStream Voicevox(string strMsg, uint intSpeaker, bool booPlay = true)
        {
            try
            {
                //MainForm me = new MainForm();
                //Core core = me.GetDefaultCoreObject();
                //core.LoadModel(intSpeaker);
                AudioQueryOptions audioQueryOptions = new AudioQueryOptions()
                {
                    kana = false
                };
                bool enableInterrogativeUpspeak = false;
                string json = core.AudioQuery(strMsg, intSpeaker, audioQueryOptions);
                SynthesisOptions synthesisOptions = new SynthesisOptions()
                {
                    enable_interrogative_upspeak = enableInterrogativeUpspeak
                };
                // When
                byte[] wavData = core.Synthesis(json, intSpeaker, synthesisOptions);
                MemoryStream ms = new MemoryStream(wavData);
                if (booPlay)
                {
                    var player = new SoundPlayer(ms);
                    player.PlaySync();
                    ms.Dispose();
                    ms = null;
                }
                return ms;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }
        }
        private async void silenceTimer_Tick(object sender, EventArgs e)
        {
            try
            {
                silenceTimerCnt++;
                //AIから話しかける仕組み
                if (silenceTimerCnt == 3)
                {
                    Random rTalk = new Random();
                    if (rTalk.Next(0, 1) == 0) //乱数で0が出たら話しかける。トリガーは"？？"
                    {
                        await Chat("？？");
                    }
                }
                if (chkMic.Checked)
                {
                    if (recognizer.State == SpeechRecognizerState.Idle)
                    {
                        await recognizer.ContinuousRecognitionSession.StartAsync();
                    }
                }
                else
                {
                    if (recognizer.State != SpeechRecognizerState.Idle)
                    {
                        await recognizer.ContinuousRecognitionSession.StopAsync();
                    }
                }

            }
            catch (Exception ex) 
            {
                string msg = ex.Message;
            }
        }
        private async void txtMsg_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Enter)
            {
                e.Handled = true;
                // ユーザーからの質問
                string strMsg = txtMsg.Text;
                txtMsg.Text = "";
                await Chat(strMsg);
                if (Sdprompt != "")
                {
                    lblInfo.Text = "Now generating images...";
                    lblInfo.Visible = true;
                    Application.DoEvents();
                    AimGenForm frm = new AimGenForm(DatabasePath);
                    string imgPath = Application.StartupPath + @$"\img\sd{DateTime.Now.ToString("yyyyMMddhhmmssfff")}.png";
                    frm.InitSetting();
                    frm.AimGenRun(Sdprompt, imgPath);
                    frm.Dispose();
                    lblInfo.Visible = false;
                }
                Sdprompt = "";
            }
        }
        private async void chkMic_CheckedChanged(object sender, EventArgs e)
        {
            try
            {
                if (chkMic.Checked)
                {
                    if (recognizer.State == SpeechRecognizerState.Idle)
                    {
                        await recognizer.ContinuousRecognitionSession.StartAsync();
                    }
                    silenceTimer.Start();
                }
                else
                {
                    if (recognizer.State != SpeechRecognizerState.Idle)
                    {
                        await recognizer.ContinuousRecognitionSession.StopAsync();
                    }
                    silenceTimer.Stop();
                }
            }
            catch (Exception ex)
            {
                if (ex.HResult == -2147199735) MessageBox.Show("Have you enabled the \"online speech recognition\" in Windows settings under Start > Settings > Privacy > Speech? And have you enabled the microphone under the app capabilites?", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                else MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                chkMic.Checked = false;
                chkMic.Enabled = false;
            }
        }
        private void btnEdit_Click(object sender, EventArgs e)
        {
            int intNo = 0;
            if (cboChatList.Items.Count > 0) intNo = int.Parse(cboChatList.SelectedValue.ToString());
            var frm = new EditChatForm(DatabasePath, intNo);
            if (frm.ShowDialog() == DialogResult.OK)
            {
                cboChatList.DataSource = mng();
                if (intNo > 0)
                {
                    cboChatList.SelectedValue = intNo;
                    if (cboChatList.SelectedIndex == -1) cboChatList.SelectedIndex = 1;
                }
            }
            if (cboChatList.Items.Count == 0) btnLoad.Enabled = false;
            else btnLoad.Enabled = true;
        }
        private DataTable mng()
        {
            try
            {
                DataTable dt = new DataTable();
                using (SQLiteConnection con = new SQLiteConnection("Data Source=" + DatabasePath))
                {
                    SQLiteDataAdapter adp = new SQLiteDataAdapter("SELECT * from mng order by no;", con);
                    adp.Fill(dt);
                }
                return dt;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }
        }
        private DataTable ch()
        {
            try
            {
                DataTable dt = new DataTable();
                SQLiteDataAdapter adp;
                using (SQLiteConnection con = new SQLiteConnection("Data Source=" + DatabasePath))
                {
                    adp = new SQLiteDataAdapter($"SELECT id,replace(replace(msg,'User:',''),'Assistant:','') FROM ch{cboChatList.SelectedValue} where sq>3 and flg=0 order by sq;", con);
                    adp.Fill(dt);
                }
                return dt;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }
        }
        private string vvx_name(uint intVvx)
        {
            try
            {
                string strRtn = "";
                var conSb = new SQLiteConnectionStringBuilder { DataSource = DatabasePath };
                var con = new SQLiteConnection(conSb.ToString());
                con.Open();
                using (var cmd = new SQLiteCommand(con))
                {
                    cmd.CommandText = $"select * from vvx where no={intVvx}";
                    using (var reader = cmd.ExecuteReader())
                    {
                        if (reader.Read())
                        {
                            strRtn = (string)reader["nam"];
                        }
                    }
                }
                con.Close();
                return strRtn;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return "";
            }
        }
        private void LoadSetting()
        {
            dtChat = ch();
            BindingSource bindsrc = new BindingSource();
            bindsrc.DataSource = dtChat;
            dvwMsg.DataSource = bindsrc;
            for (int i = 0; i < dvwMsg.Rows.Count; i++)
            {
                if ((string)dvwMsg.Rows[i].Cells[0].Value == "User") dvwMsg.Rows[i].DefaultCellStyle = dvwStyle2;
                else dvwMsg.Rows[i].DefaultCellStyle = dvwStyle1;
                dvwMsg.Rows[i].Height = 52;
            }
            dvwMsg.Columns[0].Width = 100;
            dvwMsg.Columns[0].Resizable = DataGridViewTriState.False;
            dvwMsg.Columns[0].ReadOnly = true;
            dvwMsg.Columns[1].Width = dvwMsg.Width - 130;
            dvwMsg.Columns[1].Resizable = DataGridViewTriState.False;
            dvwMsg.Columns[1].ReadOnly = true;
            dvwMsg.Columns[1].DefaultCellStyle.WrapMode = DataGridViewTriState.True;

            if (dvwMsg.Rows.Count > 0) dvwMsg.FirstDisplayedScrollingRowIndex = dvwMsg.Rows.Count - 1;

            txtMsg.Enabled = true;
            chkMic.Enabled = true;
            btnEdit.Enabled = false;
            btnSm.Enabled = false;
            btnSd.Enabled = false;
            btnLlv.Enabled = false;
            dvwMsg.AllowDrop = true;

            txtMsg.Focus();
        }
        private void WriteHistory(AuthorRole aurID, string strMsg)
        {
            try
            {
                Dictionary<string, AuthorRole>? Roles = new Dictionary<string, AuthorRole> { { "System", AuthorRole.System }, { "User", AuthorRole.User }, { "Assistant", AuthorRole.Assistant } };
                DataRow dr = dtChat.NewRow();
                dr[0] = Roles.FirstOrDefault(v => v.Value.Equals(aurID)).Key;
                dr[1] = strMsg;
                dtChat.Rows.Add(dr);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }
        private void dvwMsg_Refresh()
        {
            int rix = dvwMsg.Rows.Count - 1;
            if ((string)dvwMsg.Rows[rix].Cells[0].Value == "User") 
            { 
                dvwMsg.Rows[rix].DefaultCellStyle = dvwStyle2; 
            }
            else
            {
                dvwMsg.Rows[rix].DefaultCellStyle = dvwStyle1;
                picEmt.Image = imgEmt[SearchEmotion((string)dvwMsg.Rows[rix].Cells[1].Value)];
                Application.DoEvents();
            } 
            dvwMsg.Rows[rix].Height = 52;
            if (dvwMsg.Rows.Count > 0) dvwMsg.FirstDisplayedScrollingRowIndex = dvwMsg.Rows.Count - 1;
        }
        private void btnEject_Click(object sender, EventArgs e)
        {
            ClearItems();
            dvwMsg.DataSource = null;
            dvwMsg.Rows.Clear();
            txtMsg.Enabled = false;
            chkMic.Checked = false;
            chkMic.Enabled = false;
            lblSyt.Text = "";
            btnEdit.Enabled = true;
            btnSm.Enabled = true;
            btnSd.Enabled = true;
            btnLlv.Enabled = true;
            dvwMsg.AllowDrop = false;
            picEmt.Image = null;
            GPU_UsageMemory(); //GPU使用メモリ編集
        }
        private void txtMsg_Enter(object sender, EventArgs e)
        {
            txtMsg.SelectAll();
        }
        public Core GetDefaultCoreObject()
        {
            InitializeOptions options = new InitializeOptions(GetOpenJtalkDictDir());
            options.LoadAllModels = false;
            return new Core(options);
        }
        public string GetOpenJtalkDictDir()
        {
            return Path.Combine(GetThisAppDirectory() ?? "", "open_jtalk_dic_utf_8-1.11");
        }
        static public Assembly GetThisAssembly()
        {
            return Assembly.GetExecutingAssembly();
        }
        static public string? GetThisAppDirectory()
        {
            string appPath = GetThisAssembly().Location;
            return Path.GetDirectoryName(appPath);
        }
        private void GPU_UsageMemory()
        {
            try
            {
                var category = new PerformanceCounterCategory("GPU Adapter Memory");
                var counterNames = category.GetInstanceNames();
                var gpuCountersDedicated = new List<PerformanceCounter>();
                var gpuCountersShared = new List<PerformanceCounter>();
                var result = 0f;
                double GpuUse = 0d;
                foreach (string counterName in counterNames)
                {
                    foreach (var counter in category.GetCounters(counterName))
                    {
                        if (counter.CounterName == "Dedicated Usage")
                        {
                            gpuCountersDedicated.Add(counter);
                        }
                    }
                }
                gpuCountersDedicated.ForEach(x =>
                {
                    _ = x.NextValue();
                });
                Task.Delay(1000);
                gpuCountersDedicated.ForEach(x =>
                {
                    result += x.NextValue();
                });
                double usage = (double)result;
                GpuUse = Math.Round(usage / 107374182.4, MidpointRounding.AwayFromZero);
                GpuUse = GpuUse / 10;
                lblGpu.Text = $"{GpuUse.ToString("0.0")} GB";
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }
        private void btnSd_Click(object sender, EventArgs e)
        {
            AimGenForm frm = new(DatabasePath);
            if (frm.ShowDialog() == DialogResult.OK) Sd = true;
            frm.Dispose();
        }
        private void CheckSd()
        {
            var frm = new AimGenForm(DatabasePath);
            frm.InitSetting();
            Sd = !frm.CheckItem(false);
            frm.Dispose();
            frm = null;
        }
        private void CheckSm()
        {
            var frm = new EditSummaryForm(DatabasePath);
            frm.InitSetting();
            Sm = !frm.CheckItem(false);
            frm.Dispose();
            frm = null;
        }
        private void CheckLv()
        {
            var frm = new LlavaForm(DatabasePath);
            frm.InitSetting();
            Sm = !frm.CheckItem(false);
            frm.Dispose();
            frm = null;
        }
        private void btnSm_Click(object sender, EventArgs e)
        {
            var frm = new EditSummaryForm(DatabasePath);
            if (frm.ShowDialog() == DialogResult.OK) Sm = true;
            frm.Dispose();
            frm = null;
        }
        private void dvwMsg_CellMouseDoubleClick(object sender, DataGridViewCellMouseEventArgs e)
        {
            if(dvwMsg.Rows[e.RowIndex].Cells[0].Value.ToString() == "Assistant")
            {
                string strMsg = dvwMsg.Rows[e.RowIndex].Cells[1].Value.ToString();

                //画像生成キーワード探索
                if (strMsg.IndexOf("[")  != -1)
                {
                    var rgxRtn = new Regex(@"\[(.+?)\]").Matches(strMsg);
                    if (rgxRtn.Count > 0)
                    {
                        Sdprompt = rgxRtn[0].Value.Replace("[", "").Replace("]", "").Trim();
                        if (Sdprompt != "")
                        {
                            lblInfo.Text = "Now generating images...";
                            lblInfo.Visible = true;
                            Application.DoEvents();
                            AimGenForm frm = new AimGenForm(DatabasePath);
                            string imgPath = Application.StartupPath + @$"\img\sd{DateTime.Now.ToString("yyyyMMddhhmmssfff")}.png";
                            frm.InitSetting();
                            frm.AimGenRun(Sdprompt, imgPath);
                            frm.Dispose();
                            lblInfo.Visible = false;
                        }
                    }
                }
                else
                {
                    picEmt.Image = imgEmt[SearchEmotion((string)dvwMsg.Rows[e.RowIndex].Cells[1].Value)];
                    Application.DoEvents();
                    if (!Wt)
                    {
                        Wt = true;
                        SpeechSynthesis((string)dvwMsg.Rows[e.RowIndex].Cells[1].Value); //音声ソフト実行
                        Thread.Sleep(3000);
                        Wt = false;
                    }
                }
            }
        }
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (this.WindowState == FormWindowState.Normal)
            {
                Properties.Settings.Default.WindowSize = this.Size;
                Properties.Settings.Default.WindowLocation = this.Location;
            }
            else
            {
                Properties.Settings.Default.WindowLocation = this.RestoreBounds.Location;
                Properties.Settings.Default.WindowSize = this.RestoreBounds.Size;
            }
            Properties.Settings.Default.Save();
        }
        private void dvwMsg_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop)) e.Effect = DragDropEffects.Copy;
        }
        private async void dvwMsg_DragDrop(object sender, DragEventArgs e)
        {
            try
            {
                if (e.Data != null)
                {
                    string[] strFiles = (string[])e.Data.GetData(DataFormats.FileDrop);
                    Image img = Image.FromFile(strFiles[0]);
                    img.Dispose();
                    var llv = new Llava(DatabasePath);
                    var frm = new LlavaForm(DatabasePath);
                    frm.InitSetting();
                    if(!frm.CheckItem(false))
                    {
                        lblInfo.Text = "Now analyzing images...";
                        lblInfo.Visible = true;
                        Application.DoEvents();
                        string strMsg = await llv.Run(strFiles[0]);
                        lblInfo.Visible = false;
                        Application.DoEvents();
                        if (strMsg != "")
                            await Chat($"ユーザーは画像を見せました。{strMsg}");
                    }
                    frm.Dispose();
                    frm = null;
                    llv = null;
                }
            }
            catch (Exception ex)
            {
                string msg = ex.Message;
            }
        }

        private void btnLlv_Click(object sender, EventArgs e)
        {
            var frm = new LlavaForm(DatabasePath);
            if (frm.ShowDialog() == DialogResult.OK) Lv = true;
            frm.Dispose();
            frm = null;
        }

        private async void DiscordBot() 
        {
            if (string.IsNullOrWhiteSpace(strTkn))
            {
                Console.WriteLine("Please specify a token in the DISCORD_TOKEN environment variable.");
                Environment.Exit(1);
                return;
            }
            dcconfig = new()
            {
                Token = strTkn,
                TokenType = TokenType.Bot,
                Intents = DiscordIntents.AllUnprivileged | DiscordIntents.MessageContents | DiscordIntents.GuildVoiceStates
            };
            dcclient = new(dcconfig);
            dcclient.UseVoiceNext();

            //メッセージイベント
            dcclient.MessageCreated += OnMessageCreated;
            dcclient.Ready += OnClientReady;
            dcclient.GuildAvailable += OnGuildAvailable;
            dcclient.VoiceStateUpdated += OnVoiceStateUpdated;

            dcstatus = new("with fire", DSharpPlus.Entities.ActivityType.Playing);
            await dcclient.ConnectAsync(dcstatus, UserStatus.Online);
            picDc.Visible = true;
        }

        private async Task OnMessageCreated(DiscordClient client, MessageCreateEventArgs e)
        {
            //ボットに応答しない
            if (e.Message.Author.Id==client.CurrentUser.Id) return;
            //ターゲットチャンネル
            if (lngCid != 0 && e.Channel.Id != lngCid) return;
            if(!Wt)
            {
                Wt = true;
                ChatHistory.Message msg = new(AuthorRole.User, "User: " + e.Message.Content);
                if (Op) chtDB.WriteHistory(AuthorRole.User, "User: " +  e.Message.Content);
                WriteHistory(AuthorRole.User, e.Message.Content);
                Invoke(dvwMsg_Refresh);
                // 回答の表示
                string strMsg = "";
                await foreach (string strAns in chtSess.ChatAsync(msg, infPara)) strMsg += strAns;
                //発信するときは「User:」や「Assistant:」を抜く
                string strSndmsg = strMsg.Replace("User:", "").Replace("Assistant:", "").Replace("assistant:", "").Trim();
                await e.Message.RespondAsync(strSndmsg);
                if (Op) chtDB.WriteHistory(AuthorRole.Assistant, strMsg);
                WriteHistory(AuthorRole.Assistant, strSndmsg);
                Invoke(dvwMsg_Refresh);
                //if (dcvoiceCon is not null) //Voiceチャンネルが開いているときは音声を送信
                //{
                //    MemoryStream msRes = SpeechSynthesis(strSndmsg, false);
                //    await SendAsync(dcvoiceCon, msRes);
                //}
                Wt = false;
            }
        }

        private Task OnClientReady(DiscordClient client, ReadyEventArgs e)
        {
            Console.WriteLine("Bot is ready!");
            return Task.CompletedTask;
        }

        private Task OnGuildAvailable(DiscordClient client, GuildCreateEventArgs e)
        {
            Console.WriteLine($"Guild available: {e.Guild.Name}");
            return Task.CompletedTask;
        }

        private async Task OnVoiceStateUpdated(DiscordClient client, VoiceStateUpdateEventArgs e)
        {
            try
            {
                //Bot以外のチャンネルが開いた場合
                if (e.After.Channel != null && !e.After.User.IsBot)
                {
                    Console.WriteLine($"{e.After.User.Username} joined voice channel {e.After.Channel.Name}");

                    var vnext = client.GetVoiceNext();
                    var connection = await vnext.ConnectAsync(e.After.Channel);

                    dcvoiceCon = connection;

                }
                //Botが退出した場合はコネクションをNULLにする
                if (e.After.Channel == null && e.After.User.IsBot)
                {
                    dcvoiceCon = null;
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }
        private async Task SendAsync(VoiceNextConnection connection, MemoryStream ms)
        {
            try
            {
                var transmit = connection.GetTransmitSink();
                var pcm = ConvertAudioToPcm(ms);
                await pcm.CopyToAsync(transmit);
                await pcm.DisposeAsync();
                ms.Close();
                ms.Dispose();
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }
        private Stream ConvertAudioToPcm(MemoryStream ms)
        {
            try
            {
                var ffmpeg = Process.Start(new ProcessStartInfo
                {
                    FileName = "ffmpeg",
                    Arguments = $@"-i pipe:0 -ac 2 -f s16le -ar 48000 pipe:1",
                    RedirectStandardInput = true,
                    RedirectStandardOutput = true,
                    UseShellExecute = false
                });

                ffmpeg.StandardInput.BaseStream.WriteAsync(ms.ToArray(), 0, (int)ms.Length);
                ffmpeg.StandardInput.BaseStream.FlushAsync();
                return ffmpeg.StandardOutput.BaseStream;
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
                return null;
            }
        }

        private MemoryStream SpeechSynthesis(string strMsg, bool booPlay = true)
        {
            switch (intSyt) //音声ソフト選択
            {
                case 1:
                    return CevioAI(strMsg, strCvo, booPlay);
                case 2:
                    return Voicevox(strMsg, intVvx, booPlay);
            }
            return null;
        }

        private int SearchEmotion(string strMsg)
        {
            //Smiling
            if (Regex.IsMatch(strMsg, "😀|😃|😄|😁|😆|😅|🤣|😂|🙂|🙃|😉|😊|😇")) return 1;
            //Affection
            if (Regex.IsMatch(strMsg, "🥰|😍|🤩|😘|😗|☺|😚|😙|🥲|💌|💘|💝|💖|💗|💓|💞|💕|💟|❣|❤|🧡|💛|💚|💙|💜|🤎|🖤|🤍|💋")) return 2;
            //Skeptical
            if (Regex.IsMatch(strMsg, "🤐|🤨|😐|😑|😶|😶‍🌫️|😏|😒|🙄|😬|😮‍💨|🤥")) return 3;
            //Concerned
            if (Regex.IsMatch(strMsg, "😕|😟|🙁|☹|😮|😯|😲|😳|🥺|🥹|😦|😧|😨|😰|😥|😢|😭|😱|😖|😣|😞|😓|😩|😫|🥱")) return 4;
            //Negative
            if (Regex.IsMatch(strMsg, "😤|😡|😠|🤬|😈|👿|💀|☠")) return 5;

            return 0;
        }

    }
}
